package edu.georgetown.gucs.experiment;

import edu.georgetown.gucs.tokenizers.TokenizerParameters;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.JGraph;
import edu.georgetown.gucs.utility.Pair;

import java.io.*;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Evan B.
 * @since 6/17/14.
 */
public class Experiment {

    private ExperimentConfReader config;

    //initialized in constructor
    private int trialThreads;
    private String outputFile = null;

    private ArrayList<Trial> threads;

    private List<ExperimentResults> trialResults = new ArrayList<>();


    private boolean finished = false;

    TrialParameters poison = new TrialParameters();

    private static final Logger logger = Logger.getLogger(Experiment.class.getName());


    //  private final ResultTracker tracker = ResultTracker.getResult();


    /**
     * Constructor that takes a configFile
     *
     * @param configFile configuration options
     */
    public Experiment(String configFile) {

        config = new ExperimentConfReader(configFile);

        trialThreads = config.getNumTrialThreads();
        threads = new ArrayList<>(trialThreads);


        System.out.println("[Total Trials] " + config.getTotalTrialCount());

    }

    /**
     * public function to run trials
     * Print results to stdOutput
     * OR
     * Save results to File
     */
    public void runExperiment() {

        logger.log(Level.INFO, "[Starting trials]");

        int runNumber = runTrials();

    }


    private int runTrials() {


        logger.log(Level.INFO, "[Inside runTrials]");

        int runNumber = 0;

        FileLister fileLister = new FileLister(config.getDataSetPath().getAbsolutePath(),
                config.getNumTrialThreads());


        //logger.log(Level.INFO, "[FileLister Size] " + fileLister.size());

        // We also want to wait while the file lister runs, as it is threaded and we don't want concurrency
        // issues
        try {
            while (!fileLister.getLoaded()) {

                logger.log(Level.INFO, "[fileLister not loaded, sleep 250MS]");
                Thread.sleep(250);
            }
        } catch (InterruptedException e) {
            ExceptionHandler.handleExceptionExit(e);
            logger.log(Level.INFO, "[Interrupted Exception in fileLister]");
        }


        LinkedBlockingQueue<TrialParameters> threadParameterQueue =
                new LinkedBlockingQueue<>(trialThreads * 10);

        for (int i = 0; i < trialThreads; i++) {

            logger.log(Level.INFO, "[Starting trial threads]");
            threads.add(i, new Trial(threadParameterQueue, poison, this));

            threads.get(i).setName("Trial thread " + i);
            threads.get(i).setDaemon(true);
            threads.get(i).start();
        }


        unpackRunTrial(fileLister, threadParameterQueue);

        return runNumber;
    }

    public void unpackRunTrial(FileLister fileLister, LinkedBlockingQueue<TrialParameters> threadParameterQueue) {


        logger.log(Level.INFO, "[Unpacking Trial Params]");

        Random random = new Random(System.currentTimeMillis());

        for (Double dictionarySampleSize : config.getDictionaryParameters()) {

            logger.log(Level.INFO, "[Unpacking DictionaryParams " + dictionarySampleSize);

            for (Double sampleSize : config.getSampleParameters()) {

                logger.log(Level.INFO, "[Unpacking sample params] " + sampleSize);

                for (TokenizerParameters tokenizerParameters : config.getTokenizerParameters()) {

                    logger.log(Level.INFO, "[tokenizeParams] " + tokenizerParameters);

                    for (Pair<Double, Double> idfPair : config.getIDF()) {

                        logger.log(Level.INFO, "[Unpacking idfPair] " + idfPair);

                        for (String fingerprinter : config.getFingerprinter()) {

                            logger.log(Level.INFO, "[Unpacking fingerprinter] " + fingerprinter);

                            for (Pair<String, Integer> matcherPair : config.getMatcherVector()) {

                                logger.log(Level.INFO, "[Unpacking matcherPair] " + matcherPair.getA() + " ," +
                                        matcherPair.getB());

                                for (String splitter : tokenizerParameters.splitters()) {

                                    logger.log(Level.INFO, "[Unpacking Splitter] " + splitter);

                                    for (String mangler : tokenizerParameters.manglers()) {

                                        logger.log(Level.INFO, "[Unpacking mangler] " + mangler);
                                        //System.out.println(
                                        //      "Experiment Splitter: " + splitter + "  Mangler(s): " + mangler);


                                        for (int i = 0; i < config.getNumTrials(); i++) {

                                            logger.log(Level.INFO, "[Creating trialParameter Objects]");
                                            String matcherName = matcherPair.getA();

                                            int minScore = matcherPair.getB();

                                            TrialParameters trialParameters =
                                                    new TrialParameters(config.getDataSetDirectory(),
                                                            fileLister,
                                                            tokenizerParameters.tokenizers(),
                                                            dictionarySampleSize,
                                                            sampleSize, idfPair, fingerprinter, matcherName,
                                                            minScore, mangler, splitter);

                                            trialParameters.setRandom(random);
                                            try {

                                                threadParameterQueue.put(trialParameters);

                                                //  experimentID++;
                                                //System.err.println("!!++> Adding parameter set, current length is: " + threadParameterQueue.size());
                                            } catch (InterruptedException e) {
                                                ExceptionHandler.handleExceptionExit(e);

                                                logger.log(Level.INFO, "[ThreadPareteQueue.put] interrupted exception");
                                            }

                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        logger.log(Level.INFO, "[Exiting Unpacking]");
        try {
            threadParameterQueue.put(poison);
        } catch (InterruptedException e) {
            ExceptionHandler.handleExceptionExit(e);
        }


        // Wait for the threads to all terminate

        while (!finished) {
            try {
                Thread.sleep(10000);
            } catch (InterruptedException e) {
                ExceptionHandler.handleExceptionExit(e);
            }


            int finishedThreadCount = 0;
            for (int i = 0; i < trialThreads; i++) {
                //System.err.println("  ---> " + threads.get(i).getState());
                if (threads.get(i).getState() == Thread.State.valueOf("TERMINATED")) {
                    finishedThreadCount++;
                }
            }
            if (finishedThreadCount == trialThreads) {
                finished = true;
            }
        }

        logger.info("Experiment runTrials end");

    }

    /**
     * Change this to use a queue with thread
     *
     * @param results the result of each trial
     *                Removed synchronized
     */
    public synchronized void getResults(ExperimentResults results) {
        // Seperate thread to output results
        // Thread shuts down when main thread is killed
        // Potential memory leak
        ResultTracker.getResult().log(results);

        trialResults.add(results);

    }


    public void graphResults(String fileName) {

        JGraph graph = new JGraph(averageTrials(), fileName);
        graph.setTitle("Experiment Results");
        graph.savePlot();

    }

    /**
     * @return a list containing the average score
     * for a trial with x parameters
     */
    private HashMap<String, Pair<Integer,Double>> averageTrials() {


        HashMap<String, ArrayList<ExperimentResults>> totalResults = new HashMap<>();


        for (ExperimentResults avgResult : trialResults) {

            String key = avgResult.getMangler() + ":" + avgResult.getIdfRange().getA() + "_" +
                    avgResult.getIdfRange().getB();


            //change this to reflect IDF Range && Mangler
            if (totalResults.containsKey(key)) {

                totalResults.get(key).add(avgResult);

            } else {
                //add idfRange to the key
                ArrayList<ExperimentResults> trial = new ArrayList<>();
                trial.add(avgResult);

                totalResults.put(key, trial);
            }


        }


        HashMap<String, Pair<Integer, Double>> totalAverages = new HashMap<>();
        for (Map.Entry<String, ArrayList<ExperimentResults>> it : totalResults.entrySet()) {

            Integer temp = 0;
            for (ExperimentResults result : it.getValue()) {

                temp += result.getAverageMatcherScore();

            }
            temp /= it.getValue().size();

            double stdDeviation = stdDeviation(it.getValue(),temp);
           // System.out.println("Mean: " + temp + " Deviation+: " + (stdDeviation + temp) + " Deviation-: " + (temp - stdDeviation) );

            totalAverages.put(it.getKey(), new Pair<>(temp, stdDeviation));

        }


        return totalAverages;
    }


    public double stdDeviation(ArrayList<ExperimentResults> totalAverages, double average) {

        double temp = 0;
        for (ExperimentResults it : totalAverages) {

            temp += Math.pow(it.getAverageMatcherScore() - average, 2);

        }

        return  ( Math.sqrt(temp/totalAverages.size())/ (Math.sqrt(totalAverages.size()) * 1.96) );

    }

    public void setOutputFile(String outputFile) {

        this.outputFile = outputFile;

    }


    /**
     * Prints the output of trials to STD out
     * OR
     * Saves the output of trials to a file
     */
    public void printResults() {

        if (outputFile != null) {
            writeOutputFile(this.outputFile);

        } else {

            for (ExperimentResults result : trialResults) {

                result.printTrialParams();

            }
        }
    }

    public void printGraph() {

        ExperimentGraph expGraph = new ExperimentGraph();
        expGraph.calculateManglerScores(trialResults, "average");
        expGraph.start();

    }

    public void saveGraphCoords(String graphCordsName, String matcher) {
        HashMap<String, ArrayList<Integer>> experimentScores;

        ExperimentGraph expGraph = new ExperimentGraph();
        expGraph.calculateManglerScores(trialResults, matcher);

        experimentScores = expGraph.getAverages();

        if (graphCordsName != null) {

            File file = new File(graphCordsName);
            try (BufferedWriter output = new BufferedWriter(new FileWriter(file))) {

                for (Map.Entry<String, ArrayList<Integer>> average : experimentScores.entrySet()) {

                    output.write(String.format("%n%s %d%n",
                            average.getKey(), average.getValue().get(0)));

                }
                output.close();

            } catch (IOException e) {
                e.printStackTrace();
            }

        }
    }

    /**
     * Save a trials into a file
     *
     * @param fileName the file containing results
     */
    private void writeOutputFile(String fileName) {
        File file = new File(fileName);
        try (BufferedWriter output = new BufferedWriter(new FileWriter(file))) {

            System.out.println("writing output file");
            for (ExperimentResults result : trialResults) {

                output.write(String.format("%n%n%s%n%s%n%s%n%s%n%s%n%s%n",
                        "[Mangler] " + result.getMangler(),
                        "[Matcher Name] " + result.getMatcherName(),
                        "[Average Matcher Score] " + result.getAverageMatcherScore(),
                        "[Exact Matcher Score] " + result.getAverageExactScore(),
                        "[IDF Range] " + result.getIdfRange().getA()
                                + "-" + result.getIdfRange().getB(),
                        "[File Tokenizers]" + result.getTokenizers()));

            }

        } catch (IOException e) {
            e.printStackTrace();
        }


    }

}
