package edu.georgetown.gucs.fingerprinter;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.georgetown.gucs.tokenizers.TokenizeFile;
import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.*;

/**
 * Creates a fingerprint that consist of a bit vector representing the presence or absence of terms in a particular
 * dictionary
 *
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class BitVectorFingerprinter extends Fingerprinter {

    private static final Logger logger = Logger.getLogger(BitVectorFingerprinter.class.getName());

    /**
     * Constructor that generates the fingerprint name, version, unique identifier (GUID), system identifier, and creating
     * program. Sets the fingerprinter output to display all available information.
     */
    public BitVectorFingerprinter() {
        super();
    }

    /**
     * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
     * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
     * information.
     *
     * @param dictionaryFilename the filename containing the dictionary to use for this fingerprinter
     */
    public BitVectorFingerprinter(String dictionaryFilename) {
        super(dictionaryFilename);
    }

    /**
     * Loads a previously generated fingerprinter digest from fingerprinter and dictionary XML files. <br>
     * <br>
     * If dictionaryFilename is null, assumes the full dictionary is contained in the fingerprinter XML file or the file
     * containing the dictionary specified in the fingerprinter XML file exists in the current directory. If the full
     * dictionary is contained in the fingerprinter XML file but does not exist in the current directory, a dictionary XML
     * file containing the dictionary is created.
     *
     * @param fingerprintFilename the string name of the XML file containing the fingerprinter's digest
     * @param dictionaryFilename  the filename containing the dictionary to use for this fingerprinter
     */
    public BitVectorFingerprinter(String fingerprintFilename, String dictionaryFilename) {
        super(fingerprintFilename, dictionaryFilename);
    }

    /**
     * Computes a byte array fingerprint indicating the presence or absence of each token in this dictionary
     */
    @Override
    public List<Fingerprint> computeFingerprint(String filename) {

        List<Fingerprint> bitVectorList = new ArrayList<>();

        //initial fingerprint set  * Tokenize the input
        AntlrBitSet bitvector = new AntlrBitSet(this.dictionary.size());

        //dictionary = initial tokens
        TokenizerList tokenizers = this.dictionary.getTokenizers();

        /*
         * Tokenize the input
         *
         */
        logger.log(Level.INFO, "[Begin BVPF tokenization]");

        for (Entry<String, List<Token>> splitterEntry : tokenizers.tokenizeFile(filename).entrySet()) {

          //  System.out.println("SplitterValue " + splitterEntry.getValue().size());
            logger.log(Level.INFO, "[Current Mangler] " + splitterEntry.getKey() +
                    " [TokenList size] " + splitterEntry.getValue().size());
            if (splitterEntry.getValue().size() > 0) {

                //Create a list of the tokens that exist after the bitVectorFingerprinter
                //weeds out tokens that shouldn't exist

                //compare tokens in segment to dictionary {use bloom filter}
                for (Token token : splitterEntry.getValue()) {

                    int position = this.dictionary.getPosition(token);
                    //> -1 means the token exists so add it
                    if (position > -1) {
                        bitvector.add(position);
                    }
                }

                logger.log(Level.INFO, "[BitVectorSize] " + bitvector.size());
                int FIRST_TOKEN = 0;
                List<Token> tokenList = splitterEntry.getValue();
                bitVectorList.add(new Fingerprint(filename, bitvector.toBytes(), tokenList, tokenList.get(FIRST_TOKEN)
                        .start_byte(), tokenList.get(tokenList.size() - 1).end_byte()));
            }
        }
        logger.log(Level.INFO, "[BitVectorList size] " + bitVectorList.size());
        return bitVectorList;
    }

    @Override
    public String toString() {
        return "BitVectorFingerprinter";
    }

    /**
     * Creates a bit vector fingerprint representing the presence or absence of terms in a particular dictionary
     *
     * @param args array of string command line arguments<br>
     *             <code>args[0]</code> the filename containing the dictionary to use for this fingerprinter<br>
     *             <code>args[1]</code> the filename of the document to fingerprint<br>
     *             <code>args[2]</code> optional: the filename to write the XML output; if not specified, uses
     *             fingerprint_randomInteger.xml<br>
     *             <code>args[3]</code> optional: the filename containing the output specifications
     */

    public static void main(String args[]) {
        String dictionary_file = "";
        String input_file = "";
        String output_file = "";
        String output_config = null;

        try {
            dictionary_file = args[0];
            input_file = args[1];
            output_file = (args.length > 2) ? args[2] : "fingerprint_" + (new Random()).nextInt(Integer.MAX_VALUE) + ".xml";
            if (args.length > 3) {
                output_config = args[3];
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            System.err
                    .println("Usage: java BitVectorFingerprinter <dictionary file> <file to fingeprint> "
                            + "<optional: destination filename (otherwise use: fingerprint_<randomInteger>.xml)> <optional: output configuration file>");
            System.exit(-1);
        }

        try {
            BitVectorFingerprinter bvfp = new BitVectorFingerprinter(dictionary_file);
            if (output_config != null) {
                bvfp.outputFields(output_config);
            }
            System.out.println(bvfp.generateXML(bvfp.computeFingerprintXML(input_file), input_file, output_file));

        } catch (RuntimeException exception) {
            ExceptionHandler.handleExceptionExit(exception);
        }
    }
}
