package edu.georgetown.gucs.fingerprinter;

import edu.georgetown.gucs.tokenizers.TokenizeFile;
import edu.georgetown.gucs.tokenizers.Tokenizer;
import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.*;

import java.io.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * @author Evan Bloomberg
 * @since 4/30/14.
 */
public class SdhashFingerprinter extends Fingerprinter {

    private Map<String, List<Token>> tokenizedFileMap;
    private final Random generator = new Random(System.currentTimeMillis());
    private final String threadName = Thread.currentThread().getName().replaceAll("\\s", "");
    private Tokenizer tokenizers = null;

    StringBuilder sdHashOutput = new StringBuilder();

    private static final Logger logger = Logger.getLogger(Fingerprinter.class.getName());

    //initializes the score threshold to 0
    public SdhashFingerprinter() {
        super();
    }

    /**
     * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
     * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
     * information.
     *
     * @param dictionaryFilename the filename containing the dictionary to use for this fingerprinter
     */
    public SdhashFingerprinter(String dictionaryFilename) {
        super(dictionaryFilename);
        List<String> tokenizerNames = this.dictionary.getTokenizerNames();
        tokenizers = new TokenizerList(tokenizerNames);
    }

    /**
     * Returns a random number
     *
     * @return a random double based on @value generators value
     */
    private String randomGenerator() {
        return String.valueOf(generator.nextDouble() * 0.5);
    }

    /**
     * Tokenizes a file based on a list of tokenizers that are passed to the function
     *
     * @param newTokenizers a list of tokenizers that will be used to tokenize the file
     */
    public void setTokenizers(TokenizerList newTokenizers) {

        tokenizers = new TokenizerList(newTokenizers.getNames());
    }

    /**
     * SetTokenizers sets the tokenizer configuration for the input file
     *
     * @param tokenizerConfigFileName filename of the tokenizer configuration
     */
    public void setTokenizers(String tokenizerConfigFileName) {
        tokenizers = new TokenizeFile(tokenizerConfigFileName);
    }


    /**
     * computeFingerprint tokenizes a file in order to create a fingerprint
     *
     * @param fileName the file to tokenize
     * @return a list of fingerprints
     */
    @Override
    public List<Fingerprint> computeFingerprint(String fileName) {

        List<Fingerprint> fingerprint = new ArrayList<>();


        if (tokenizers != null) {
            tokenizedFileMap.put(fileName, tokenizers.tokenize(fileName));

        } else {

            System.err.println("Tokenizers == NULL");
            //throw an exception if tokenizedFileMap has not been set
        }

        StringBuilder tokenContainer = new StringBuilder();
        for (Map.Entry<String, List<Token>> entry : tokenizedFileMap.entrySet()) {


            for (Token token : entry.getValue()) {

                //does it matter that this isn't a sychronized List?

                tokenContainer.append(token).append(' ');
                System.out.println(tokenContainer);
            }
        }

        try {

            //stash fingerprint in temp file
            File tempFile = File.createTempFile("sdhashFpr" + threadName + randomGenerator(), null);
            tempFile.deleteOnExit();

            try (BufferedWriter bufferedWriter = new BufferedWriter(new FileWriter(tempFile))) {
                bufferedWriter.write(tokenContainer.toString());
            }


            try (InputStream inStream = new ProcessBuilder(Global.sdHashFilePath, tempFile.getCanonicalPath())
                    .redirectErrorStream(true).start().getInputStream();
                 BufferedReader buffer = new BufferedReader(new InputStreamReader(inStream, Global.fileEncoding))) {

                String line;
                while ((line = buffer.readLine()) != null) {
                    sdHashOutput.append(line);
                }

                byte[] fingerprintByte = sdHashOutput.toString().getBytes();
                fingerprint.add(new Fingerprint(fileName, fingerprintByte, 0, fingerprintByte.length));
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

        return fingerprint;
    }

    public void printSdHashOutput(){

        System.out.println(sdHashOutput.toString());

    }

}
