package edu.georgetown.gucs.matcher;

import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.utility.AntlrBitSet;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;

/**
 * Provides a score based on a cosine similarity match to be used with a {@linkplain BitVectorFingerprinter}. Uses the
 * {@linkplain AntlrBitSet} to do the calculations.
 * 
 * @author Clay Shields
 **/
public class CosineSimilarityFingerprintMatcher extends FingerprintMatcher {

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints to zero
   **/
  public CosineSimilarityFingerprintMatcher() {
    super();
  }

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints
   * 
   * @param score
   *          the minimum score needed to be considered a match
   */
  public CosineSimilarityFingerprintMatcher(int score) {
    super(score);
  }

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints. The string parameter is needed for
   * use with ObjectBuilder.
   * 
   * @param score
   *          the string value of the minimum score needed to be considered a match
   */
  public CosineSimilarityFingerprintMatcher(String score) {
    super(Integer.valueOf(score));
  }

  /**
   * Determines that the two fingerprints are matching if their similarity score is at or above the minimum score for
   * this fingerprinter
   */
  @Override
  public boolean match(byte[] fp1, byte[] fp2) {
      return getScore(fp1, fp2) >= this.minimum_score;
  }

  /**
   * Determines the cosine similarity score for these two fingerprints
   */
  @Override
  public int getScore(byte[] fp1, byte[] fp2) {

    if (fp1 == null || fp2 == null) {
      ExceptionHandler.handleExceptionThrow(new Exception("Null fingerprints given for matching"));
    }

    // Convert the byte arrays to BitSets

    AntlrBitSet matchingBits = new AntlrBitSet(fp1);
    AntlrBitSet secondBitset = new AntlrBitSet(fp2);
    int firstDegree = matchingBits.degree();
    int secondDegree = secondBitset.degree();

    // If either degree of the arrays is zero, return zero
    // @evan swapped 0 for Global.MIN_SCORE
    if (firstDegree == Global.MIN_SCORE || secondDegree == Global.MIN_SCORE) {
      return Global.MIN_SCORE;
    }

    // Find matching bits
    matchingBits.andInPlace(secondBitset);
    return (int) ((matchingBits.degree() / Math.sqrt(firstDegree * secondDegree)) * Global.MAX_SCORE);
  }
  
  @Override
  public String toString() {
    return "CosineSimilarityFingerprintMatcher";
  }

  /**
   * Provides a score based on a cosine similarity match; uses a bitVectorFingerprinter
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename containing the dictionary to use for fingerprinting<br>
   *          <code>args[1]</code> the filename of the document to fingerprint<br>
   *          <code>args[2]</code> the filename of the document to fingerprint
   */
  public static void main(String[] args) {
    String dictionary = null;
    String file1 = null;
    String file2 = null;
    try {
      dictionary = args[0];
      file1 = args[1];
      file2 = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CosineSimilarityFingerprintMatcher <dictionary> <file to fingerprint> <file to fingerprint>");
      System.exit(-1);
    }

    // Create two fingerprinters
    BitVectorFingerprinter fingerprinter = new BitVectorFingerprinter(dictionary);
    CosineSimilarityFingerprintMatcher matcher = new CosineSimilarityFingerprintMatcher();
    for (Fingerprint fingerprint_outer : fingerprinter.computeFingerprintXML(file1)) {
      for (Fingerprint fingerprint_inner : fingerprinter.computeFingerprintXML(file2)) {
        System.out.println(fingerprint_outer.getFileName());
        System.out.println(fingerprint_inner.getFileName());
        System.out.println("Score: " + matcher.getScoreXML(fingerprint_outer.getBase64Fingerprint(), fingerprint_inner.getBase64Fingerprint()));
        System.out.println();
      }
    }
  }
}
