package edu.georgetown.gucs.matcher;

import java.util.Arrays;

import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.utility.Global;

/**
 * Determines if two fingerprints' documents are the same; uses a {@linkplain BitVectorFingerprinter}
 * 
 * @author Clay Shields
 */
public class ExactFingerprintMatcher extends FingerprintMatcher {

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints to zero
   **/
  public ExactFingerprintMatcher() {
    super();
  }

  /**
   * Determines that the two fingerprints are matching if their byte arrays are equal
   */
  @Override
  public boolean match(byte[] fp1, byte[] fp2) {
    return Arrays.equals(fp1, fp2);
  }

  /**
   * Determines a similarity score for these two fingerprints
   * 
   * @return the similarity score between these two fingerprints; returns 0 if the fingerprints are not an exact match,
   *         returns 99 they are
   */
  @Override
  public int getScore(byte[] fp1, byte[] fp2) {
    if (match(fp1, fp2))
      return Global.MAX_SCORE;
    return Global.MIN_SCORE;
  }
  
  @Override
  public String toString() {
    return "ExactFingerprintMatcher";
  }

  /**
   * Determines if two fingerprints' documents are the same; uses a bitVectorFingerprinter
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename containing the dictionary to use for fingerprinting<br>
   *          <code>args[1]</code> the filename of the document to fingerprint<br>
   *          <code>args[2]</code> the filename of the document to fingerprint
   */
  public static void main(String[] args) {
    String dictionary = null;
    String file1 = null;
    String file2 = null;

    try {
      dictionary = args[0];
      file1 = args[1];
      file2 = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java ExactFingerprintMatcher <dictionary> <fingeprint file> <fingeprint file>");
      System.exit(-1);
    }

    // Create two fingerprinters
    BitVectorFingerprinter fingerprinter = new BitVectorFingerprinter(dictionary);
    ExactFingerprintMatcher matcher = new ExactFingerprintMatcher();
    for (Fingerprint fingerprint_outer: fingerprinter.computeFingerprintXML(file1)) {
      for (Fingerprint fingerprint_inner : fingerprinter.computeFingerprintXML(file2)) {
        System.out.println(fingerprint_outer.getFileName());
        System.out.println(fingerprint_inner.getFileName());

        if (matcher.match(fingerprint_outer.getBase64Fingerprint(), fingerprint_inner.getBase64Fingerprint())) {
          System.out.println("MATCH");
        } else {
          System.out.println("NOT A MATCH");
        }
        System.out.println();
      }
    }
  }
}
