package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.Dictionary;
import java.util.List;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

import edu.georgetown.gucs.utility.Global;
import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.cn.ChineseAnalyzer;
import org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer;
import org.apache.lucene.analysis.tokenattributes.OffsetAttribute;
import org.apache.lucene.analysis.tokenattributes.TermAttribute;
import org.apache.lucene.util.Version;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Token;

/**
 * Splits contents of a Chinese or Chinese-English text file into tokens using Apache Lucene's
 * {@linkplain ChineseAnalyzer} or {@linkplain SmartChineseAnalyzer}. Tokens are used for {@linkplain Dictionary}
 * creation.
 *
 * @author Lindsay Neubauer
 */
public class ChineseFileTokenizer extends FileTokenizer {

    private static final long serialVersionUID = 3508035370868840628L;

    /**
     * Constructor that sets the token creation mode to split by smart tokenization using probabilistic word segmentation
     *
     * @see org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer
     */
    public ChineseFileTokenizer() {
        this.mode = "smart";
    }

    /**
     * Constructor that sets the token creation mode. If set to "individual" the file is split by one Chinese character as
     * one Chinese word; if set to "smart" the file is split by smart tokenization using probabilistic word segmentation.
     *
     * @param mode
     *          the string mode to split tokens by one Chinese character as one word or by smart tokenization using
     *          probabilistic word segmentation
     */
    public ChineseFileTokenizer(String mode) {
        this.mode = mode.trim();
    }

    /**
     * Sets the token creation mode. If set to "individual" the file is split by one Chinese character as one Chinese
     * word; if set to "smart" the file is split by smart tokenization using probabilistic word segmentation.
     *
     * @param mode
     *          the string mode to split tokens by one Chinese character as one word or by smart tokenization using
     *          probabilistic word segmentation
     */
    @Override
    public void setMode(String mode) {
        this.mode = mode.trim();
    }

    /**
     * Splits a Chinese or mixed Chinese-English document into tokens based on the token creation mode. If mode is set to
     * "individual" the file is split by one Chinese character as one Chinese word; if set to "smart" the file is split by
     * smart tokenization using probabilistic word segmentation. Assumes the file is UTF-8 encoded and Version.LUCENE_30
     * for "smart" tokenization.
     *
     * @param filename
     *          the string filename of the document to split into tokens
     *
     * @see org.apache.lucene.analysis.cn.ChineseAnalyzer
     * @see org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer
     */
    @Override
    public List<Token> tokenize(String filename) {
        if (this.mode.equalsIgnoreCase("individual")) {
            return individual_tokenize(filename);
        }
        return smart_tokenize(filename);
    }

    /**
     * Splits the document into tokens by one Chinese character as one Chinese word. Numeric tokens are removed. English
     * tokens larger than 1 character are included. Assumes the file is UTF-8 encoded.
     *
     * @param filename
     *          the string filename of the document to split into tokens
     *
     * @return the vector of tokens present in the file
     *
     * @see org.apache.lucene.analysis.cn.ChineseAnalyzer
     */
    private List<Token> individual_tokenize(String filename) {

        try {
            BufferedReader buffer_reader = (filename.endsWith(".gz")) ? new BufferedReader(new InputStreamReader(
                    new GZIPInputStream(new FileInputStream(filename)), Global.fileEncoding)) : new BufferedReader(new InputStreamReader(
                    new FileInputStream(filename), Global.fileEncoding));
            ChineseAnalyzer chineseAnalyzer = new ChineseAnalyzer();
            TokenStream tokenStream = chineseAnalyzer.tokenStream(null, buffer_reader);
            TermAttribute termAttribute = tokenStream.addAttribute(TermAttribute.class);
            OffsetAttribute offsetAttribute = tokenStream.addAttribute(OffsetAttribute.class);
            tokenStream.reset();

            synchronized (this.tokenVectorMap) {
                this.tokenVectorMap.clear();
                List<Token> tokenVector = Collections.synchronizedList(new Vector<Token>());
                synchronized (tokenVector) {
                    while (tokenStream.incrementToken()) {
                        tokenVector
                                .add(new Token(termAttribute.term(), offsetAttribute.startOffset(), offsetAttribute.endOffset()));
                    }
                    tokenStream.end();
                    tokenStream.close();
                    buffer_reader.close();
                    this.tokenVectorMap.put("null", tokenVector);
                    return tokenVector;
                }
            }
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
        return null;
    }

    /**
     * Splits document into tokens. Uses probabilistic knowledge to find the optimal word segmentation for Simplified
     * Chinese text. The text is first broken into sentences, then each sentence is segmented into words. Assumes the file
     * is UTF-8 encoded and Version.LUCENE_30 for the SmartChineseAnalyzer.
     *
     * @param filename
     *          the string filename of the document to split into tokens
     *
     * @return the vector of tokens present in the file
     *
     * @see org.apache.lucene.analysis.cn.smart.SmartChineseAnalyzer
     */
    private List<Token> smart_tokenize(String filename) {

        try {
            BufferedReader buffer_reader = (filename.endsWith(".gz")) ? new BufferedReader(new InputStreamReader(
                    new GZIPInputStream(new FileInputStream(filename)), Global.fileEncoding)) : new BufferedReader(new InputStreamReader(
                    new FileInputStream(filename), Global.fileEncoding));
            SmartChineseAnalyzer smartChineseAnalyzer = new SmartChineseAnalyzer(Version.LUCENE_30);
            TokenStream tokenStream = smartChineseAnalyzer.tokenStream(null, buffer_reader);
            TermAttribute termAttribute = tokenStream.addAttribute(TermAttribute.class);
            OffsetAttribute offsetAttribute = tokenStream.addAttribute(OffsetAttribute.class);
            tokenStream.reset();

            synchronized (this.tokenVectorMap) {
                this.tokenVectorMap.clear();
                List<Token> tokenVector = Collections.synchronizedList(new Vector<Token>());
                synchronized (tokenVector) {
                    while (tokenStream.incrementToken()) {
                        tokenVector
                                .add(new Token(termAttribute.term(), offsetAttribute.startOffset(), offsetAttribute.endOffset()));
                    }

                    tokenStream.end();
                    tokenStream.close();
                    buffer_reader.close();
                    this.tokenVectorMap.put("null", tokenVector);
                    return tokenVector;
                }
            }
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
        return null;
    }
}
