package edu.georgetown.gucs.tokenizers;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Vector;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;

/**
 * Tests tokenizers using a configuration file that contains a list of {@linkplain Tokenizer} objects to use on a
 * document and prints the results of the tokenization
 *
 * @author Clay Shields
 *
 */
public class TokenizeFile extends Tokenizer {

    private static final long serialVersionUID = -1504621044845334628L;
    private List<String> tokenizers;
    private String splitter = null;

    /**
     * Constructor that initializes an empty string vector of tokenizers to use
     */
    public TokenizeFile() {
        this.tokenizers = new Vector<>();
    }

    /**
     * Constructor that sets the tokenizers from a configuration file to use on this document. The first tokenizer in the
     * configuration file must support reading from a file.
     *
     * @param config_file
     *          the string XML filename that specifies these tokenizers
     */
    public TokenizeFile(String config_file) {
        this.tokenizers = new Vector<>();
        addTokenizers(config_file);
    }

    /**
     * Adds tokenizers from an XML file to the list of tokenizers to use on this document
     *
     * @param config_file
     *          the string XML filename that contains the tokenizers
     */
    @SuppressWarnings("unchecked")
    private void addTokenizers(String config_file) {
        try {
            File xml = new File(config_file);
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(xml);
            Element dict = doc.getRootElement();
            Element splitterElement;
            for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
                String tokenizer = tokenizerElement.getAttributeValue("name");
                if (this.tokenizers.size() == 0) {
                    if (!tokenizer.contains("FileTokenizer")) {
                        ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: First tokenizer (" + tokenizer
                                + ") must read from a file."));
                    } else if ((splitterElement = tokenizerElement.getChild("splitter")) != null) {
                        this.splitter = splitterElement.getAttributeValue("name") + " "
                                + splitterElement.getAttributeValue("value");
                    }
                }

                if (tokenizerElement.getAttributeValue("parameter") != null)
                    tokenizer += " " + (tokenizerElement.getAttributeValue("parameter"));
                this.tokenizers.add(tokenizer.trim());
            }
        } catch (JDOMException | IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
    }

    /**
     * Adds tokenizers from the given list to the list of tokenizers to use on this document
     *
     * @param tokenizerNames
     *          the list of tokenizer names to use
     */
    public void addTokenizers(List<String> tokenizerNames) {
        this.tokenizers = tokenizerNames;
    }

    /**
     * Splits the given file into tokens and alters or eliminates those tokens based on the vector of tokenizers. Prints
     * each token to standard output. The first tokenizer in the vector of tokenizers must support reading from a file.
     *
     * @param filename
     *          the name of the file to be tokenized
     */
    public void tokenizeFile(String filename) {
        Global.monitor.writeToBuffer("TokenizeFile", "tokenize", "tokenizers=" + this.tokenizers.toString());
        TokenizerList tokenizerList = new TokenizerList(this.tokenizers);

       // System.out.println(tokenizerList.toString());

        tokenizerList.setSplitter(this.splitter);
        this.tokenVectorMap = tokenizerList.tokenizeFile(filename);
    }

    /**
     * Tokenizes a text file and prints the resulting tokens to the the screen
     *
     * @param args
     *          array of string command line arguments<br>
     *          <code>args[0]</code> the filename of the document to split into tokens<br>
     *          <code>args[1]</code> the filename of the tokenizer configuration file
     */
    public static void main(String args[]) {
        String input_file = "";
        String config_file = "";
        try {
            input_file = args[0];
            config_file = args[1];
        } catch (ArrayIndexOutOfBoundsException e) {
            System.err.println("Usage: java -jar TokenizeFile.jar <file name> <tokenizers config file>");
            System.err.println("Tokenizes a text file and prints the resulting tokens to the the screen");
            System.exit(-1);
        }

        try {
            TokenizeFile tokenizeFile = new TokenizeFile(config_file);
            tokenizeFile.tokenizeFile(input_file);
            tokenizeFile.printTokens();
            Global.monitor.closeBuffer();
        } catch (RuntimeException exception) {
            ExceptionHandler.handleExceptionExit(exception);
        }
    }
}
