package edu.georgetown.gucs.utility;

import java.util.concurrent.Executor;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.Semaphore;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Limits the number of threads to run at one time when running more tasks than there are cores
 *
 * @author <a href="http://stackoverflow.com/questions/2001086/how-to-make-threadpoolexecutors-submit-method-block-if-it-is-saturated">How to make ThreadPoolExecutor's submit() method block if it is saturated?</a>
 */
public class BoundedExecutor {

    /**
     * the semaphore that keeps a count of the number of available threads and acts accordingly
     */
    final Semaphore semaphore;
    private final Executor exec;

    private static final Logger logger = Logger.getLogger(BoundedExecutor.class.getName());

    /**
     * Constructor that sets the executor and semaphore upper bound
     *
     * @param exec  the executor object that submits runnable tasks
     * @param bound the <code>int<code> maximum number of permits available for the semaphore; this value
     *              may be negative, in which case releases must occur before any acquires will be granted
     */
    public BoundedExecutor(Executor exec, int bound) {

        logger.log(Level.INFO, "bound:", bound);
        this.exec = exec;
        this.semaphore = new Semaphore(bound);
    }

    /**
     * Calls <code>run</code> method of given instance; blocks until a thread is available
     *
     * @param instance the instance of runnable
     */
    public void submitTask(final Runnable instance) {
        try {
            this.semaphore.acquire();
            this.exec.execute(new Runnable() {
                @Override
                public void run() {
                    try {
                        instance.run();
                    } finally {
                        BoundedExecutor.this.semaphore.release();
                    }
                }
            });
        } catch (RejectedExecutionException e) {
            this.semaphore.release();
            ExceptionHandler.handleExceptionExit(e);
        } catch (InterruptedException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
    }
}
