package edu.georgetown.gucs.utility;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Prints exception messages and stack traces before calling <code>System.exit(-1)</code>
 *
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class ExceptionHandler {

    private static final Logger logger = Logger.getLogger(ExceptionHandler.class.getName());

    /**
     * Prints the message and stack trace for this Exception and exits; used with built in java exceptions (not
     * SDText-specific)
     *
     * @param exception the exception that provides information on error
     */
    public static void handleExceptionExit(Exception exception) throws RuntimeException {
        System.err.println("ExceptionHandler:" + exception.getMessage());
        System.err.println("ExceptionHandler " + Arrays.toString(exception.getStackTrace()));
        System.err.println("ExceptionHandler: " + exception.getMessage());
        logger.log(Level.SEVERE, Arrays.toString(exception.getStackTrace()));
        throw new RuntimeException(exception.getMessage());
        //System.exit(-1);
    }

    /**
     * Prints the message and stack trace for this Exception and throws a runtime exception; used with exceptions that are
     * SDText-specific (to give user feedback about error)
     *
     * @param exception the exception that provides information on error
     */
    public static void handleExceptionThrow(Exception exception) {
        logger.log(Level.SEVERE, exception.getMessage());
        throw new RuntimeException(exception.getMessage());
    }

    /**
     * Prints the message, SQLState, vendor-specific exception code, and stack trace for this SQLException
     *
     * @param exception the sqlException that provides information on the database access error or other errors
     */
    public static void handleSQLException(SQLException exception) {
        SQLException sql_exception = exception;
        do {
            String exceptionMessage = String.format("%n%s %n%s %n%s %n%s%n", "SQL Exception:", sql_exception.getMessage(),
                    "ANSI-92 SQL State: " + sql_exception.getSQLState(), "Vendor Error Code: " + sql_exception.getErrorCode());
            logger.log(Level.SEVERE, exceptionMessage);
            logger.log(Level.SEVERE,  Arrays.toString(sql_exception.getStackTrace()));
            logger.log(Level.SEVERE,  exceptionMessage);
            logger.log(Level.SEVERE, Arrays.toString(sql_exception.getStackTrace()));
            sql_exception = sql_exception.getNextException();
        } while (sql_exception != null);
        throw new RuntimeException(exception.getMessage());
        //System.exit(-1);

    }

    /**
     * Prints the message, SQLState, vendor-specific exception code, and stack trace for this SQLException
     *
     * @param exception the sqlException that provides information on the database access error or other errors
     * @param database  the connection for the database that caused this exception
     * @param statement the SQL statement that caused this exception
     */
    public static void handleSQLException(SQLException exception, Connection database, Statement statement) {
        SQLException sql_exception = exception;
        do {
            String stmt = (statement != null) ? String.format("%nStatement: %s", statement.toString()) : "";
            String exceptionMessage = String.format("%n%s %n%s%s %n%s %n%s%n", "SQL Exception:", sql_exception.getMessage(),
                    stmt, "ANSI-92 SQL State: " + sql_exception.getSQLState(), "Vendor Error Code: " + sql_exception.getErrorCode());
            logger.log(Level.SEVERE,  exceptionMessage);
            logger.log(Level.SEVERE,  Arrays.toString(sql_exception.getStackTrace()));
            logger.log(Level.SEVERE, "ExceptionHandler: " + exceptionMessage);
            logger.log(Level.SEVERE, Arrays.toString(sql_exception.getStackTrace()));
            sql_exception = sql_exception.getNextException();
        } while (sql_exception != null);

        try {
            if (statement != null) {
                statement.close();
            }
            if (database != null) {
                database.close();
            }
        } catch (SQLException ex) {
            System.err.println("ExceptionHandler: "  +
                    "Error closing statement or database in ExceptionHandler.handleSQLException");
            handleSQLException(ex);
        }
        Global.monitor.closeBuffer();
        throw new RuntimeException(exception.getMessage());
        //System.exit(-1);
    }
}
