package sdtext;

import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Vector;

import javax.swing.DefaultListModel;
import javax.swing.GroupLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.ListSelectionModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.TransferHandler;

import edu.georgetown.gucs.utility.Global;

/**
 * Creates a handler for a drag and drop selection; used with {@linkplain sdtext.GUI}.
 * 
 * @see <a href="http://www.java-forums.org/awt-swing/3030-help-drag-drop-jlist.html">Help with drag and drop on
 *      JList</a>
 * 
 * @author Lindsay Neubauer
 * 
 */
class ExperimentSelectionHandler extends TransferHandler {

  private static final long serialVersionUID = 8433300841181809587L;
  DataFlavor localArrayListFlavor;
  final DataFlavor serialArrayListFlavor;
    private JList source = null;
  private int[] indices = null;
  private int itemIndex = -1; // Location of item added
  private int itemsAdded = 0; // Number of items added

  private final String enumName;
  final Vector<String> selectedVector = new Vector<String>();
  final DefaultListModel optionModel;
  final DefaultListModel selectionModel;
  private final JLabel optionLabel = new JLabel();
  private final JLabel selectedLabel = new JLabel();
  private JScrollPane optionScrollPane;
  private JScrollPane selectedScrollPane;

  /**
   * Constructor that sets the appropriate dataFlavor
   * 
   * @param enumName
   *          string name of the enum to use for this selection handler
   */
  public ExperimentSelectionHandler(String enumName) {
    try {
        String localArrayListType = DataFlavor.javaJVMLocalObjectMimeType + ";class=java.util.ArrayList";
        this.localArrayListFlavor = new DataFlavor(localArrayListType);
    } catch (ClassNotFoundException e) {
      Global.monitor.printErrorMessage("ExperimentSelectionHandler", "Unable to create data flavor");
      Global.monitor.writeToBuffer("ExperimentSelectionHandler", "ExperimentSelectionHandler",
          "Unable to create data flavor");
    }
    this.serialArrayListFlavor = new DataFlavor(ArrayList.class, "ArrayList");
    this.optionModel = new DefaultListModel();
    this.selectionModel = new DefaultListModel();
    this.enumName = enumName;
  }

  /**
   * Creates the panel that contains tokenizer selection
   * 
   * @param layout
   *          the groupLayout to use for the tokenizer panel
   */
  protected void createPanel(GroupLayout layout) {
    int height = 10;
    if (this.enumName.equals("FingerprinterEnum")) {
      height += 20 * FingerprinterEnum.values().length;
      for (FingerprinterEnum matcherEnum : FingerprinterEnum.values()) {
        this.optionModel.add(0, matcherEnum.name);
        this.optionLabel.setText("Fingerprinter Options");
        this.selectedLabel.setText("Selected Fingerprinters");
      }
    } else if (this.enumName.equals("MatcherEnum")) {
      height += 20 * MatcherEnum.values().length;
      for (MatcherEnum matcherEnum : MatcherEnum.values()) {
        this.optionModel.add(0, matcherEnum.name);
        this.optionLabel.setText("Matcher Options");
        this.selectedLabel.setText("Selected Matchers");
      }
    } else if (this.enumName.equals("ManglerEnum")) {
      height += 18 * ManglerEnum.values().length;
      for (ManglerEnum manglerEnum : ManglerEnum.values()) {
        this.optionModel.addElement(manglerEnum.name);
        this.optionLabel.setText("Mangler Options");
        this.selectedLabel.setText("Selected Manglers");
      }
    }
    this.optionScrollPane = addLists("Options", this.optionModel);
    this.selectedScrollPane = addLists("Selected", this.selectionModel);
    this.selectedScrollPane.setPreferredSize(new Dimension(250, height));
    this.optionScrollPane.setPreferredSize(this.selectedScrollPane.getPreferredSize());
    layout.setHorizontalGroup(layout.createSequentialGroup()
        .addGroup(layout.createParallelGroup().addComponent(this.optionLabel).addComponent(this.optionScrollPane))
        .addGroup(layout.createParallelGroup().addComponent(this.selectedLabel).addComponent(this.selectedScrollPane)));
    layout
        .setVerticalGroup(layout
            .createParallelGroup()
            .addGroup(layout.createSequentialGroup().addComponent(this.optionLabel).addComponent(this.optionScrollPane))
            .addGroup(
                layout.createSequentialGroup().addComponent(this.selectedLabel).addComponent(this.selectedScrollPane)));
    layout.linkSize(SwingConstants.VERTICAL, this.optionLabel, this.selectedLabel);
    layout.linkSize(SwingConstants.VERTICAL, this.optionScrollPane, this.selectedScrollPane);
  }

  /**
   * Provides the components for mangler selection
   * 
   * @return the array of JComponents for mangler selection
   */
  protected JComponent[] getManglers() {
    return new JComponent[] { this.optionLabel, this.optionScrollPane, this.selectedLabel, this.selectedScrollPane };
  }

  /**
   * Provides the selected items
   * 
   * @return the string vector created from the selected items
   */
  protected Vector<String> getSelected() {
    return this.selectedVector;
  }

  private JScrollPane addLists(String listName, DefaultListModel tokenizerModel) {
    JList list = new JList(tokenizerModel);
    list.setName(listName);
    list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    list.setTransferHandler(this);
    list.setDragEnabled(true);
    return new JScrollPane(list);
  }

  @SuppressWarnings("unchecked")
  @Override
  public boolean importData(JComponent target, Transferable transferable) {
    ArrayList<String> alist = null;
    if (!canImport(target, transferable.getTransferDataFlavors())) {
      return false;
    }
    try {
      if (hasLocalArrayListFlavor(transferable.getTransferDataFlavors())) {
        alist = (ArrayList<String>) transferable.getTransferData(this.localArrayListFlavor);
      } else if (hasSerialArrayListFlavor(transferable.getTransferDataFlavors())) {
        alist = (ArrayList<String>) transferable.getTransferData(this.serialArrayListFlavor);
      } else {
        return false;
      }
    } catch (UnsupportedFlavorException ufe) {
      Global.monitor.printErrorMessage("ExperimentSelectionHandler", "Unsupported data flavor");
      Global.monitor.writeToBuffer("ExperimentSelectionHandler", "importData", "Unsupported data flavor");
      return false;
    } catch (IOException ioe) {
      Global.monitor.printErrorMessage("ExperimentSelectionHandler", "I/O exception");
      Global.monitor.writeToBuffer("ExperimentSelectionHandler", "importData", "I/O exception");
      return false;
    }

    if (target.getName().equals("Selected") && target != this.source && this.enumName.equals("MatcherEnum")) {
      MatcherEnum currentMatcher = MatcherEnum.valueOf(((String) this.source.getSelectedValue()).split("\\s")[0]);
      if (currentMatcher.name.contains("CosineSimilarityFingerprintMatcher")) {
        JSpinner spinner = new JSpinner(new SpinnerNumberModel(0, 0, 99, 1));
        JPanel spinnerPanel = new JPanel(new GridLayout(2, 1));
        spinnerPanel.add(new JLabel("Minimum CosineSimiarlity score to be considered a match: "));
        spinnerPanel.add(spinner);
        int retVal = JOptionPane.showConfirmDialog(target, spinnerPanel, "Minimum Score", JOptionPane.DEFAULT_OPTION,
            JOptionPane.QUESTION_MESSAGE);
        if (retVal == JOptionPane.NO_OPTION || retVal == JOptionPane.CLOSED_OPTION) {
          return false;
        }
        alist.set(alist.indexOf(currentMatcher.name), currentMatcher.name + " " + String.valueOf(spinner.getValue()));
      }
    } else if (target.getName().equals("Selected") && target != this.source && this.enumName.equals("ManglerEnum")) {
      ManglerEnum currentMangler = ManglerEnum.valueOf(((String) this.source.getSelectedValue()).split("\\s")[0]);
      if (!currentMangler.name.equals("Null")) {
        JSpinner spinner = new JSpinner(new SpinnerNumberModel(10, 1, 99, 1));
        JPanel spinnerPanel = new JPanel(new GridLayout(2, 1));
        spinnerPanel.add(new JLabel(currentMangler.name + ":   " + currentMangler.description));
        JPanel percentPanel = new JPanel(new GridLayout(1, 2));
        percentPanel.add(new JLabel("Percent Chance: "));
        percentPanel.add(spinner);
        spinnerPanel.add(percentPanel);
        int retVal = JOptionPane.showConfirmDialog(target, spinnerPanel, currentMangler.name + " Percent",
            JOptionPane.DEFAULT_OPTION, JOptionPane.QUESTION_MESSAGE);
        if (retVal == JOptionPane.NO_OPTION || retVal == JOptionPane.CLOSED_OPTION) {
          return false;
        }
        alist.set(alist.indexOf(currentMangler.name), currentMangler.name + " " + String.valueOf(spinner.getValue()));
      }
    }

    if (target instanceof JList) {
      int index = ((JList) target).getSelectedIndex();
      // Dropping the same data on itself and has no effect
      if (this.source.equals(target)) {
        if (this.source.getName().equals("Options")) {
          return false;
        } else if (this.indices != null && index >= this.indices[0] - 1
            && index <= this.indices[this.indices.length - 1]) {
          this.indices = null;
          return true;
        }
      }

      DefaultListModel listModel = (DefaultListModel) ((JList) target).getModel();
      int max = listModel.getSize();
      if (index < 0) {
        index = max;
      } else {
        index++;
        if (index > max) {
          index = max;
        }
      }
      this.itemIndex = index;
      this.itemsAdded = alist.size();
      for (Object selectedObject : alist) {
        String selectedName = (String) selectedObject;
        if (target.getName().equals("Options")) {
          this.selectedVector.remove(selectedName);
          selectedName = selectedName.split("\\s")[0];
        }
        if (this.itemsAdded > 0) {
          if (target.getName().equals("Selected")) {
            if (this.selectedVector.contains(selectedName)) {
              this.selectedVector.remove(selectedName);
            }
            this.selectedVector.add(selectedName);
          }
          listModel.add(index++, selectedName);
        }
      }
      return true;
    }
    return false;
  }

  @Override
  protected void exportDone(JComponent component, Transferable transferable, int action) {
    if (this.indices != null) {
      DefaultListModel model = (DefaultListModel) this.source.getModel();

      if (this.itemsAdded > 0) {
        if (this.source.getName().equals("SelectedTokenizers")) {
          for (int i = 0; i < this.indices.length; i++) {
            // If moving items around in the same list, need to adjust the indices accordingly
            if (this.indices[i] > this.itemIndex && this.indices[i] + this.itemsAdded < model.getSize()) {
              this.indices[i] += this.itemsAdded;
            }
          }
        }
        for (int i = this.indices.length - 1; i >= 0; i--) {
          if (!model.get(this.indices[i]).equals("CosineSimilarityFingerprintMatcher")) {
            model.remove(this.indices[i]);
          } 
        }
      }
    }
    this.indices = null;
    this.itemIndex = -1;
    this.itemsAdded = 0;
  }

  private boolean hasLocalArrayListFlavor(DataFlavor[] flavors) {
    if (this.localArrayListFlavor == null) {
      return false;
    }

    for (DataFlavor flavor : flavors) {
      if (flavor.equals(this.localArrayListFlavor)) {
        return true;
      }
    }
    return false;
  }

  private boolean hasSerialArrayListFlavor(DataFlavor[] flavors) {
    if (this.serialArrayListFlavor == null) {
      return false;
    }

    for (DataFlavor flavor : flavors) {
      if (flavor.equals(this.serialArrayListFlavor)) {
        return true;
      }
    }
    return false;
  }

  @Override
  public boolean canImport(JComponent c, DataFlavor[] flavors) {
    if (hasLocalArrayListFlavor(flavors)) {
      return true;
    }
    if (hasSerialArrayListFlavor(flavors)) {
      return true;
    }
    return false;
  }

  @Override
  protected Transferable createTransferable(JComponent component) {
    if (component instanceof JList) {
      this.source = (JList) component;
      this.indices = this.source.getSelectedIndices();
      Object[] values = this.source.getSelectedValues();
      if (values.length == 0) {
        return null;
      }
      ArrayList<String> alist = new ArrayList<String>(values.length);
      for (Object tokenizer : values) {
        alist.add(tokenizer.toString());
      }
      return new TokenizerListTransferable(alist);
    }
    return null;
  }

  @Override
  public int getSourceActions(JComponent c) {
    return MOVE;
  }

  private class TokenizerListTransferable implements Transferable {
    private final ArrayList<String> data;

    TokenizerListTransferable(ArrayList<String> alist) {
      this.data = alist;
    }

    @Override
    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException {
      if (!isDataFlavorSupported(flavor)) {
        throw new UnsupportedFlavorException(flavor);
      }
      return this.data;
    }

    @Override
    public DataFlavor[] getTransferDataFlavors() {
      return new DataFlavor[] { ExperimentSelectionHandler.this.localArrayListFlavor,
          ExperimentSelectionHandler.this.serialArrayListFlavor };
    }

    @Override
    public boolean isDataFlavorSupported(DataFlavor flavor) {
      if (ExperimentSelectionHandler.this.localArrayListFlavor.equals(flavor)) {
        return true;
      }
      if (ExperimentSelectionHandler.this.serialArrayListFlavor.equals(flavor)) {
        return true;
      }
      return false;
    }
  }
}
